# -*- coding: utf-8 -*-
import logging

import click
from path import path

from trailblazer.add import add_cmd
from trailblazer.store import api

log = logging.getLogger(__name__)


@click.command()
@click.argument('cust_dirs', type=click.Path(exists=True), nargs=-1)
@click.pass_context
def scan(context, cust_dirs):
    """Scan directory(ies) for analyses."""
    for cust_dir in cust_dirs:
        log.debug("scanning customer dir: %s", cust_dir)
        sampleinfo_files = scan_dir(cust_dir)
        for sampleinfo in sampleinfo_files:
            log.debug("adding analysis: %s", sampleinfo)
            with open(sampleinfo, 'r') as stream:
                context.invoke(add_cmd, qcsampleinfo=stream)

    api.track_update()
    context.obj['manager'].commit()


def scan_dir(root_dir):
    """Scan root for MIP analyses.

    Will look for qc sample info files.
    """
    files = path(root_dir).glob('**/**/**/*_sampleInfo.yaml')
    return files



# -*- coding: utf-8 -*-
import logging

import click
from path import path

from .restart import update_maxgaussian, restart_mip
from .start import start_mip
from trailblazer.store import api

log = logging.getLogger(__name__)


@click.group()
@click.pass_context
def analyze(context):
    pass


@analyze.command()
@click.option('-p', '--ccp', type=click.Path(exists=True), required=True)
@click.option('-a', '--analysis-type', default='genomes')
@click.option('-f', '--family', required=True)
@click.option('-c', '--config', type=click.Path(exists=True))
@click.option('-x', '--executable', type=click.Path(exists=True))
@click.option('-i', '--customer', required=True)
@click.option('-g', '--gene-list')
@click.option('-e', '--email', help='email to send errors to')
@click.option('--dryrun', is_flag=True)
@click.option('-o', '--out', type=click.File('w'), default='-')
@click.option('--conda-env')
@click.option('--script-dir', type=click.Path(exists=True))
@click.pass_context
def start(context, ccp, analysis_type, family, config, customer, gene_list,
          dryrun, executable, out, conda_env, email, script_dir):
    """Start a new analysis."""
    config = config or context.obj['mip_config']
    executable = executable or context.obj['mip_exe']
    gene_list = gene_list or context.obj['mip_genelist']
    conda_env = conda_env or context.obj.get('conda_env')
    script_dir = script_dir or context.obj.get('script_dir')

    ccp_abs = path(ccp).abspath()
    script = start_mip(
        analysis_type,
        family,
        config,
        ccp_abs,
        executable=executable,
        customer=customer,
        gene_list=gene_list,
        dryrun=dryrun,
        conda_env=conda_env,
        email=email)

    if script_dir:
        case_id = "{}-{}".format(customer, family)
        out_filename = "{}.sh".format(case_id)
        out_path = path(script_dir).joinpath(out_filename)
        click.echo(script, file=out_path.open('w'))
    else:
        click.echo(script, file=out)


@analyze.group()
@click.pass_context
def restart(context):
    pass


@restart.command('max-gaussian')
@click.option('--restart/--no-restart', default=True)
@click.option('-e', '--email', help='email to send errors to')
@click.option('-c', '--case', help='restart analysis in database')
@click.argument('config_path', type=click.Path(exists=True), required=False)
@click.pass_context
def max_gaussian(context, restart, case, email, config_path):
    """Update config file to restart with Max Gaussian for SNV enabled."""
    if case:
        most_recent = api.case(case).first()
        config_path = most_recent.config_path

    update_maxgaussian(config_path)
    log.info("updated config: {}".format(config_path))
    script_dir = context.obj.get('script_dir')
    if restart and script_dir:
        conda_env = context.obj.get('conda_env')
        kwargs = dict(executable=context.obj.get('mip_exe'), email=email,
                      conda_env=conda_env)
        restart_mip(script_dir, config_path, **kwargs)



# -*- coding: utf-8 -*-
from path import path

from trailblazer.store import Analysis
from trailblazer.exc import MissingFileError
from .sacct import parse_sacct, get_analysistime, filter_jobs
from .utils import FINISHED_STATUSES


def build_entry(sampleinfo, sacct_stream=None):
    """Build analysis status entry from sampleinfo data.

    Doesn't require any database connection.
    """
    metadata = parse_sampleinfo(sampleinfo)
    if sacct_stream:
        sacct_jobs = parse_sacct(sacct_stream)
    else:
        # try with automatically detected file
        sacct_path = path(metadata['sacct_path'])
        if not sacct_path.exists():
            raise MissingFileError(sacct_path)
        with sacct_path.open('r') as sacct_stream:
            sacct_jobs = parse_sacct(sacct_stream)
    status = determine_status(metadata['analysis_status'], sacct_jobs)
    metadata.update(status)
    new_entry = Analysis(**metadata)
    return new_entry


def determine_status(analysis_status, sacct_jobs):
    """Determine the status of an analysis."""
    if analysis_status in FINISHED_STATUSES:
        # calculate the total runtime
        # finished status should override all else;
        # do our best to figure out runtime etc.
        success_jobs = filter_jobs(sacct_jobs, failed=False)
        runtime, cputime, completed_at = get_analysistime(success_jobs)
        status = dict(status='completed', runtime=runtime, cputime=cputime,
                      completed_at=completed_at)
    else:
        non_success = filter_jobs(sacct_jobs, failed=True)
        if len(non_success) == 0:
            status = dict(status='running')
        else:
            failed_step = non_success[0]['name']
            failed_at = non_success[0]['end']
            status = dict(status='failed', failed_step=failed_step,
                          failed_at=failed_at)
    return status


def parse_sampleinfo(sampleinfo):
    """Extract basic info from loaded sampleinfo file.

    Args:
        sampleinfo (dict): loaded contents from sampleinfo file

    Returns:
        dict: extracted and transformed values from sampleinfo
    """
    fam_key = sampleinfo.keys()[0]
    fam_data = sampleinfo[fam_key][fam_key]
    sample_ids = sampleinfo[fam_key].keys()
    sample_ids.remove(fam_key)
    analysis_start = fam_data['AnalysisDate']
    analysis_type = fam_data['AnalysisType']
    analysis_out = path(fam_data['Program']['QCCollect']['OutDirectory'])
    customer = fam_data['InstanceTag'][0]
    case_id = "{}-{}".format(customer, fam_key)
    cluster_const = analysis_out.parent.parent.parent
    config_path = cluster_const.joinpath(analysis_type, fam_key,
                                         "{}_config.yaml".format(fam_key))
    sacct_path = "{}.status".format(fam_data['lastLogFilePath'])
    values = {
        'case_id': case_id,
        'pipeline': 'mip',
        'pipeline_version': fam_data['MIPVersion'],
        'type': analysis_type,
        'root_dir': analysis_out.parent.parent,
        'started_at': analysis_start,
        'config_path': config_path,
        'samples': sample_ids,
        'analysis_status': fam_data['AnalysisRunStatus'],
        'sacct_path': sacct_path,
    }
    return values



# -*- coding: utf-8 -*-
from path import path
import yaml

from .start import start_mip


def restart_mip(script_dir, config_path, **start_kwargs):
    """Restart a MIP run."""
    with open(config_path, 'r') as stream:
        values = yaml.load(stream)
    script_content = start_mip(analysis_config=config_path, **start_kwargs)
    customer = values['instanceTag'][0]
    family = values['familyID']
    out_filename = "{}-{}.sh".format(customer, family)
    out_path = path(script_dir).joinpath(out_filename)
    with out_path.open('wb') as out_stream:
        out_stream.write(script_content)


def update_maxgaussian(config_path):
    """Update MIP config to prepare for restart after max gaussian error."""
    with open(config_path, 'r') as stream:
        values = yaml.load(stream)

    # update the requried values
    # set Max Gaussian flag for SNVs
    values['GATKVariantReCalibrationSnvMaxGaussians'] = 1
    # update early (successful) programs to dry run
    early_programs = ['pBwaMem', 'pChanjoSexCheck', 'pFastQC', 'pFreebayes',
                      'pGATKBaseRecalibration', 'pGATKGenoTypeGVCFs',
                      'pGATKHaploTypeCaller', 'pGATKRealigner', 'pGZipFastq',
                      'pManta', 'pPicardToolsCalculateHSMetrics',
                      'pPicardToolsCollectMultipleMetrics',
                      'pPicardToolsMarkduplicates',
                      'pPicardToolsMergeSamFiles',
                      'pSamToolsMpileUp', 'pSambambaDepth']
    for program_key in early_programs:
        #assert values[program_key] in (1, '1')
        values[program_key] = 2

    with open(config_path, 'w') as out_stream:
        yaml_out = yaml.dump(values, default_flow_style=False)
        yaml_out_correct = yaml_out.replace('\n- ', '\n    - ')
        out_stream.write(yaml_out_correct)



# -*- coding: utf-8 -*-
import logging

from trailblazer.store import api
from .utils import same_entry

log = logging.getLogger(__name__)


def commit_analysis(manager, new_entry):
    """Store new analysis in the database."""
    # look up previous analyses for the same case
    old_entry = api.case(case_id=new_entry.case_id).first()
    if old_entry is None or not same_entry(old_entry, new_entry):
        # save the new entry to the database
        if old_entry and old_entry.status == 'running':
            # replace if status was 'running'
            log.debug("deleting existing entry: %s", new_entry.case_id)
            old_entry.delete()
            manager.commit()

        manager.add_commit(new_entry)
        log.info("added new entry: {entry.case_id} - {entry.status}"
                 .format(entry=new_entry))



# -*- coding: utf-8 -*-
from datetime import datetime
import logging

from alchy import Manager

from .models import Analysis, Model, Metadata

log = logging.getLogger(__name__)


def connect(uri):
    log.debug('open connection to database: %s', uri)
    manager = Manager(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=Model)
    return manager


def case(case_id):
    """Return analysis entries with for a case."""
    query = (Analysis.query.filter_by(case_id=case_id)
                           .order_by(Analysis.logged_at.desc()))
    return query


def analyses(analysis_id=None, since=None, is_ready=False, status=None):
    """List added analyses."""
    query = Analysis.query.order_by(Analysis.logged_at.desc())

    if since:
        log.debug("filter entries on date: %s", since)
        query = query.filter(Analysis.started_at > since)

    if analysis_id:
        log.debug("filter entries on id pattern: %s", analysis_id)
        query = query.filter(Analysis.case_id.contains(analysis_id))

    if status:
        log.debug("filter entries on status category: %s", status)
        query = query.filter_by(status=status)

    if is_ready:
        query = query.filter_by(status='completed', is_deleted=False)

    return query


def track_update():
    """Update metadata record with new updated date."""
    metadata = Metadata.query.first()
    if metadata:
        metadata.updated_at = datetime.now()



# -*- coding: utf-8 -*-
import logging

import click
import yaml

from trailblazer.exc import MissingFileError
from .commit import commit_analysis
from .core import build_entry
from .utils import is_latest_mip

log = logging.getLogger(__name__)


@click.command('add')
@click.option('-s', '--sacct', type=click.File('r'))
@click.argument('qcsampleinfo', type=click.File('r'))
@click.pass_context
def add_cmd(context, sacct, qcsampleinfo):
    """Add an analysis to the database."""
    manager = context.obj['manager']
    sampleinfo_data = yaml.load(qcsampleinfo)
    if is_latest_mip(sampleinfo_data):
        try:
            new_entry = build_entry(sampleinfo_data, sacct_stream=sacct)
            log.debug("entry: %s - %s", new_entry.case_id, new_entry.status)
            commit_analysis(manager, new_entry)
        except MissingFileError as error:
            log.error("missing file: %s", error.message)
    else:
        family_id = sampleinfo_data.keys()[0]
        log.debug("analysis version not supported: %s", family_id)



# -*- coding: utf-8 -*-
"""
#!/bin/bash -l

set -e

{% if conda_env %}
source activate {{ conda_env }}
{% endif %}

{{ command }}
"""
from datetime import datetime
import logging

from jinja2 import Template

from trailblazer.store import Analysis

log = logging.getLogger(__name__)


def build_pending(case_id, root_dir, seq_type):
    """Create an entry for an analysis which is pending."""
    new_entry = Analysis(
        case_id=case_id,
        pipeline='mip',
        started_at=datetime.now(),
        status='pending',
        root_dir=root_dir,
        type=seq_type,
    )
    return new_entry


def start_mip(analysis_type=None, family_id=None, config=None, ccp=None,
              customer=None, gene_list=None, dryrun=False, executable=None,
              conda_env=None, email=None, analysis_config=None):
    """Start a new analysis for a family.

    Args:
        analysis_type (str): 'exomes' or 'genomes'
        family_id (str): identifier for the family
        config (path): path to the MIP config file
        ccp (path): cluster constant path to the root of the analysis
        customer (Optional[str]): customer identifier (instance tag)
        gene_list (Optional[str]): name of gene list in 'references' dir
        conda_env (Optional[str]): conda environment to source
        email (Optional[str]): email to send error mails to

    Returns:
        int: return code from the executed process
    """
    if analysis_type:
        assert analysis_type in ('exomes', 'genomes'), 'invalid analysis type'
    if analysis_config is None:
        assert all([analysis_type, family_id, config, ccp])
    command = []

    # configure executable
    command.append('perl')
    command.append(executable or 'mip.pl')

    if analysis_config:
        # optionally add analysis config
        command.append('--config')
        command.append(analysis_config)

    if config:
        # add YAML config path
        command.append('--configFile')
        command.append(config)

    if analysis_type:
        # type of analysis
        command.append('--analysisType')
        command.append(analysis_type)

    if ccp:
        # configure the "cluster constant path"
        command.append('--clusterConstantPath')
        command.append(ccp)

    if customer:
        # customer (optional)
        command.append('--instanceTag')
        command.append(customer)

    if family_id:
        # add family option
        command.append('--familyID')
        command.append(family_id)

    if gene_list:
        command.append('--vcfParserSelectFile')
        command.append(gene_list)

    if dryrun:
        command.append('--dryRunAll')
        command.append('2')

    if email:
        command.append('--email')
        command.append(email)

    log.info("command: %s", ' '.join(command))
    template = Template(__doc__)
    script = template.render(command=' '.join(command), conda_env=conda_env)
    return script



# -*- coding: utf-8 -*-
FINISHED_STATUSES = ('Finished', 'Archived', 'Archiving')


def same_entry(entry, other_entry):
    """Compare two analysis entries if they are the same."""
    for aentry in [entry, other_entry]:
        if aentry.failed_step is None:
            aentry.failed_step = 'na'

    if entry.case_id != other_entry.case_id:
        return False
    elif entry.started_at != other_entry.started_at:
        return False
    elif entry.status != other_entry.status:
        return False
    elif entry.failed_step != other_entry.failed_step:
        return False
    elif entry.failed_at != other_entry.failed_at:
        return False
    else:
        return True


def is_latest_mip(sampleinfo):
    """Check if the analysis is up to date."""
    fam_key = sampleinfo.keys()[0]
    version = sampleinfo[fam_key][fam_key].get('MIPVersion')
    return version is not None and version.startswith('v3.')



# -*- coding: utf-8 -*-
from __future__ import division
import datetime
import json

import alchy
from sqlalchemy import Column, types, UniqueConstraint

STATUS_OPTIONS = ('pending', 'running', 'completed', 'failed', 'error')


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime.datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError('Type not serializable')


class JsonModel(alchy.ModelBase):

    def to_json(self, pretty=False):
        """Serialize Model to JSON."""
        kwargs = dict(indent=4, sort_keys=True) if pretty else dict()
        return json.dumps(self.to_dict(), default=json_serial, **kwargs)


Model = alchy.make_declarative_base(Base=JsonModel)


class Metadata(Model):

    """Keep track of meta data."""

    id = Column(types.Integer, primary_key=True)
    created_at = Column(types.DateTime, default=datetime.datetime.now)
    updated_at = Column(types.DateTime)


class Analysis(Model):

    """Analysis record."""

    __table_args__ = (UniqueConstraint('case_id', 'started_at', 'status',
                                       'failed_step',
                                       name='_uc_case_start_status_step'),)

    id = Column(types.Integer, primary_key=True)
    case_id = Column(types.String(128))

    # metadata
    pipeline = Column(types.Enum('mip'))
    pipeline_version = Column(types.String(32))
    logged_at = Column(types.DateTime, default=datetime.datetime.now)
    started_at = Column(types.DateTime)
    completed_at = Column(types.DateTime)
    runtime = Column(types.Integer)
    cputime = Column(types.Integer)
    status = Column(types.Enum(*STATUS_OPTIONS))
    root_dir = Column(types.Text)
    config_path = Column(types.Text)
    type = Column(types.Enum('exomes', 'genomes'))
    failed_step = Column(types.String(128), default='na')
    failed_at = Column(types.DateTime)
    comment = Column(types.Text)
    is_deleted = Column(types.Boolean, default=False)
    _samples = Column(types.Text)

    @property
    def samples(self):
        return self._samples.split(',') if self._samples else []

    @samples.setter
    def samples(self, sample_list):
        """Serialize a list of sample ids."""
        self._samples = ','.join(sample_list)

    @property
    def runtime_obj(self):
        """Format runtime as datetime object."""
        return datetime.timedelta(seconds=self.runtime)

    def failed_variantrecal(self):
        """Check if analysis failed on variat recalibration."""
        return (self.failed_step and
                'GATKVariantRecalibration' in self.failed_step)



# -*- coding: utf-8 -*-
from datetime import datetime


def filter_jobs(sacct_jobs, failed=True):
    """Filter jobs that have a FAILED etc. status."""
    if failed:
        categories = ('FAILED', 'CANCELLED')
    else:
        categories = ('COMPLETED', 'RUNNING', 'PENDING')
    filtered_jobs = [job for job in sacct_jobs if job['state'] in categories]
    return filtered_jobs


def time_to_sec(time_str):
    """Convert time in string format to seconds."""
    total_sec = 0
    if '-' in time_str:
        # parse out days
        days, time_str = time_str.split('-')
        total_sec += (int(days) * 60 * 60 * 24)

    time_parts = map(lambda val: int(round(float(val))), time_str.split(':'))
    total_sec += time_parts[-1]               # seconds
    total_sec += time_parts[-2] * 60          # minutes
    if len(time_parts) == 3:
        total_sec += time_parts[-3] * 60 * 60  # hours
    return total_sec


def convert_job(row):
    """Convert sacct row to dict."""
    state = row[9]
    if state not in ('PENDING', 'CANCELLED'):
        start_time = datetime.strptime(row[7], '%Y-%m-%dT%H:%M:%S')
        if state != 'RUNNING':
            end_time = datetime.strptime(row[8], '%Y-%m-%dT%H:%M:%S')
        else:
            end_time = None
    else:
        start_time = end_time = None

    return {
        'id': row[0],
        'name': row[1],
        'state': state,
        'start': start_time,
        'end': end_time,
        'elapsed': time_to_sec(row[6]),
        'cpu': time_to_sec(row[5]),
    }


def parse_sacct(sacct_stream):
    """Parse out information from sacct status output."""
    rows = (line.split() for line in sacct_stream)
    # filter rows that begin with a SLURM job id
    relevant_rows = (row for row in rows if row[0].isdigit())
    jobs = [convert_job(row) for row in relevant_rows]
    return jobs


def get_analysistime(sacct_jobs):
    """Calculate the total/overall time of the analysis."""
    completed_at = sacct_jobs[-1]['end']
    delta = completed_at - sacct_jobs[0]['start']
    total_cpu = sum(job['cpu'] for job in sacct_jobs)
    return int(delta.total_seconds()), total_cpu, completed_at
